<?php

namespace Drupal\crossword;

use Drupal\file\FileInterface;

/**
 * Provides an interface for crossword file parser.
 */
interface CrosswordFileParserPluginInterface {

  /**
   * Determine whether the parser works for the input file.
   *
   * @param \Drupal\file\FileInterface $file
   *   The user input file to test against the plugins.
   *
   * @return bool
   *   Returns TRUE if the plugin can parse the input file.
   */
  public static function isApplicable(FileInterface $file);

  /**
   * Returns the data array representing the parsed crossword file.
   *
   * The "data" array is what the field formatter ends up using. If you extend
   * this base class, you definitely need to override this function.
   *
   * See crossword/tests/files/test.json for example json.
   *
   * Array(
   *   'id' => $this->file->id(),
   *   'title' => 'Awesome Puzzle',
   *   'author' => 'Dan',
   *   'notepad' => 'These are notes included in the file',
   *   'puzzle' => [
   *     'grid' => array of squares
   *     'clues' => [
   *       'across' => array of clues
   *       'down' => array of clues
   *     ],
   *   ],
   * )
   *
   * A square looks like this...
   * array(
   *  'fill' => NULL or a string,
   *  'numeral' => NULL or a number,
   *  'across' => [
   *    'index' => index of across clue
   *  ],
   *  'down' => [
   *    'index' => index of down clue
   *  ],
   *  'moves' => [
   *    'up' => ['row': number, 'col': number] or NULL
   *    'down' => ['row': number, 'col': number] or NULL
   *    'left' => ['row': number, 'col': number] or NULL
   *    'right' => ['row': number, 'col': number] or NULL
   *  ],
   *  'circle' => bool,
   *  'rebus' => bool,
   * )
   *
   * Moves get added in the crossword data service.
   *
   * A clue looks like this...
   * array(
   *  'text' => string,
   *  'numeral' => number,
   *  'references' => array(
   *    [
   *      'dir' => 'down' or 'across',
   *      'numeral' => number,
   *      'index' => number,
   *    ],
   *   )
   * )
   *
   * References get added in the crossword data service.
   *
   * @return array
   *   An associative array that represents the crossword.
   */
  public function parse();

}
