<?php

namespace Drupal\crossword\Plugin\Field\FieldFormatter;

use Drupal\Component\Utility\Html;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\crossword\CrosswordDataServiceInterface;
use Drupal\file\Plugin\Field\FieldFormatter\FileFormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'crossword' formatter.
 *
 * @FieldFormatter(
 *   id = "crossword",
 *   label = @Translation("Crossword Puzzle"),
 *   weight = "1",
 *   field_types = {
 *     "crossword"
 *   }
 * )
 */
class CrosswordFormatter extends FileFormatterBase {

  /**
   * Crossword data service.
   *
   * @var \Drupal\crossword\CrosswordDataServiceInterface
   */
  protected $crosswordDataService;

  const BUTTONS = [
    'cheat' => 'Cheat',
    'solution' => 'Solution',
    'clear' => 'Clear',
    'undo' => 'Undo',
    'redo' => 'Redo',
    'instructions' => 'Instructions',
  ];

  /**
   * Constructs a FileDownloadLink object.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param \Drupal\crossword\CrosswordDataServiceInterface $crossword_data_service
   *   Token service.
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, CrosswordDataServiceInterface $crossword_data_service) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->crosswordDataService = $crossword_data_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('crossword.data_service')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    $options['print'] = TRUE;
    $options['congrats'] = 'Well done!';
    $options['details'] = [
      'title_tag' => 'h1',
      'author_tag' => 'h2',
      'notepad_tag' => 'p',
    ];
    $options['buttons']['class'] = NULL;
    $options['buttons']['buttons'] = [];
    foreach (CrosswordFormatter::BUTTONS as $key => $label) {
      switch ($key) {
        case 'clear':
          $confirm = 'Do you really want to clear? This action cannot be undone.';
          break;
        case 'solution':
          $confirm = 'Do you really want to give up?';
          break;
        default:
          $confirm = NULL;
      }
      $options['buttons']['buttons'][$key] = [
        'show' => TRUE,
        'input_label' => $label,
        'confirm' => $confirm,
      ];
    }
    $options['clues'] = [
      'do_not_render' => FALSE,
      'show' => FALSE,
      'checked' => FALSE,
      'input_label' => 'Show Clues',
    ];
    $options['errors'] = [
      'show' => TRUE,
      'checked' => FALSE,
      'input_label' => 'Show Errors',
    ];
    $options['references'] = [
      'show' => TRUE,
      'checked' => TRUE,
      'input_label' => 'Show References',
    ];
    $options['rebus'] = [
      'show' => FALSE,
      'input_label' => 'Rebus entry active',
    ];
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form = parent::settingsForm($form, $form_state);

    $field_name = $this->fieldDefinition->getName();

    $tag_options = [
      'h1' => 'h1',
      'h2' => 'h2',
      'h3' => 'h3',
      'h4' => 'h4',
      'p' => 'p',
      'div' => 'div',
      'span' => 'span',
    ];

    $form['print'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Print Styles'),
      '#default_value' => $this->getSetting('print'),
      '#description' => $this->t('Include the print stylesheet from the Crossword module.'),
    ];
    $form['congrats'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Congratulatory Message'),
      '#default_value' => $this->getSetting('congrats'),
      '#description' => $this->t('Plaintext message to display when the puzzle is solved successfully.'),
    ];
    $form['details'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Crossword Details'),
      '#tree' => TRUE,
    ];
    $form['details']['title_tag'] = [
      '#type' => 'select',
      '#title' => $this->t('Title'),
      '#default_value' => $this->getSetting('details')['title_tag'],
      '#options' => $tag_options,
      '#empty_option' => $this->t("Do not render the title"),
      '#prefix' => '<p>' . $this->t('Select the html tag to use for rendering these details.') . '</p>',
    ];
    $form['details']['author_tag'] = [
      '#type' => 'select',
      '#title' => $this->t('Author'),
      '#default_value' => $this->getSetting('details')['author_tag'],
      '#options' => $tag_options,
      '#empty_option' => $this->t("Do not render the author"),
    ];
    $form['details']['notepad_tag'] = [
      '#type' => 'select',
      '#title' => $this->t('Notepad'),
      '#default_value' => $this->getSetting('details')['notepad_tag'],
      '#options' => $tag_options,
      '#empty_option' => $this->t("Do not render the notepad"),
    ];

    if (!empty($this->getSetting('buttons'))) {
      $form['buttons'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Buttons'),
        '#tree' => TRUE,
      ];
      $form['buttons']['class'] = [
        '#type' => 'textfield',
        '#title' => $this->t('CSS Classes'),
        '#default_value' => $this->getSetting('buttons')['class'],
        '#description' => $this->t('Enter space-separated CSS classes to be added to the buttons.'),
      ];
      foreach (get_class($this)::BUTTONS as $key => $label) {
        $form['buttons']['buttons'][$key] = [
          '#type' => 'fieldset',
          '#title' => $this->t($label),
          '#tree' => TRUE,
          'show' => [
            '#type' => 'checkbox',
            '#title' => $this->t('Display Button'),
            '#default_value' => $this->getSetting('buttons')['buttons'][$key]['show'],
          ],
          'input_label' => [
            '#type' => 'textfield',
            '#title' => $this->t('Button Text'),
            '#default_value' => $this->getSetting('buttons')['buttons'][$key]['input_label'],
            '#states' => [
              'visible' => [
                ":input[name='fields[$field_name][settings_edit_form][settings][buttons][buttons][$key][show]']" => ['checked' => TRUE],
              ],
            ],
          ],
          'confirm' => [
            '#type' => 'textfield',
            '#title' => $this->t('Confirmation Message'),
            '#default_value' => $this->getSetting('buttons')['buttons'][$key]['confirm'],
            '#description' => $this->t('If empty, confirmation will not be required.'),
            '#states' => [
              'visible' => [
                ":input[name='fields[$field_name][settings_edit_form][settings][buttons][buttons][$key][show]']" => ['checked' => TRUE],
              ],
            ],
          ],
        ];
      }
    }
    if (!empty($this->getSetting('clues'))) {
      $form['clues'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Show Clues checkbox'),
        '#tree' => TRUE,
        'do_not_render' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Do Not Render Clues'),
          '#default_value' => $this->getSetting('clues')['do_not_render'],
          '#description' => $this->t('If checked, clues will only be visible in the Active Clues container. The clue lists will not be rendered anywhere on the page, and there will be no button to toggle the visibility of the clue lists.'),
        ],
        'show' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Display Checkbox'),
          '#default_value' => $this->getSetting('clues')['show'],
          '#states' => [
            'visible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][clues][do_not_render]']" => ['checked' => FALSE],
            ],
          ],
        ],
        'checked' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Check by default'),
          '#default_value' => $this->getSetting('clues')['checked'],
          '#states' => [
            'visible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][clues][show]']" => ['checked' => TRUE],
            ],
            'invisible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][clues][do_not_render]']" => ['checked' => TRUE],
            ],
          ],
        ],
        'input_label' => [
          '#type' => 'textfield',
          '#title' => $this->t('Label'),
          '#default_value' => $this->getSetting('clues')['input_label'],
          '#states' => [
            'visible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][clues][show]']" => ['checked' => TRUE],
              ":input[name='fields[$field_name][settings_edit_form][settings][clues][do_not_render]']" => ['checked' => FALSE],
            ],
          ],
        ],
      ];
    }
    if (!empty($this->getSetting('errors'))) {
      $form['errors'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Show Errors checkbox'),
        '#tree' => TRUE,
        'show' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Display Checkbox'),
          '#default_value' => $this->getSetting('errors')['show'],
        ],
        'checked' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Check by default'),
          '#default_value' => $this->getSetting('errors')['checked'],
          '#states' => [
            'visible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][errors][show]']" => ['checked' => TRUE],
            ],
          ],
        ],
        'input_label' => [
          '#type' => 'textfield',
          '#title' => $this->t('Label'),
          '#default_value' => $this->getSetting('errors')['input_label'],
          '#states' => [
            'visible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][errors][show]']" => ['checked' => TRUE],
            ],
          ],
        ],
      ];
    }
    if (!empty($this->getSetting('references'))) {
      $form['references'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Show References checkbox'),
        '#tree' => TRUE,
        'show' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Display Checkbox'),
          '#default_value' => $this->getSetting('references')['show'],
        ],
        'checked' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Check by default'),
          '#default_value' => $this->getSetting('references')['checked'],
          '#states' => [
            'visible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][references][show]']" => ['checked' => TRUE],
            ],
          ],
        ],
        'input_label' => [
          '#type' => 'textfield',
          '#title' => $this->t('Label'),
          '#default_value' => $this->getSetting('references')['input_label'],
          '#states' => [
            'visible' => [
              ":input[name='fields[$field_name][settings_edit_form][settings][references][show]']" => ['checked' => TRUE],
            ],
          ],
        ],
      ];
    }
    if (!empty($this->getSetting('rebus'))) {
      $form['rebus'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Rebus Entry checkbox'),
        '#tree' => TRUE,
        'show' => [
          '#type' => 'checkbox',
          '#title' => $this->t('Always Display Checkbox'),
          '#default_value' => $this->getSetting('rebus')['show'],
          '#description' => $this->t('If checked, the rebus entry button will always appear with the puzzle. If un-checked, the rebus entry checkbox will only be displayed on rebus puzzles.'),
        ],
        'input_label' => [
          '#type' => 'textfield',
          '#title' => $this->t('Label'),
          '#default_value' => $this->getSetting('rebus')['input_label'],
        ],
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $elements = [];
    foreach ($this->getEntitiesToView($items, $langcode) as $delta => $file) {

      $data = $this->crosswordDataService->getData($file);

      $elements[$delta] = [
        '#theme' => 'crossword',
        '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
        '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
        '#content' => [
          'title' => $this->getTitle($data),
          'author' => $this->getAuthor($data),
          'notepad' => $this->getNotepad($data),
          'across' => $this->getAcross($data),
          'down' => $this->getDown($data),
          'grid' => $this->getGrid($data),
          'controls' => $this->getControls($file),
          'active_clue' => [
            '#theme' => 'crossword_active_clues',
            '#congrats' => $this->getSetting('congrats'),
            '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
          ],
        ],
        '#attached' => [
          'library' => [
            'crossword/crossword.default',
            $this->getSetting('print') ? 'crossword/crossword.print' : '',
          ],
          'drupalSettings' => [
            'crossword' => [
              'data' => $data,
              'selector' => '.crossword',
              'congrats' => $this->getSetting('congrats') ?? NULL,
            ],
          ],
        ],
        '#attributes' => [
          'class' => [
            $this->getSetting('errors')['show'] && $this->getSetting('errors')['checked'] ? 'show-errors' : '',
            $this->getSetting('references')['show'] && $this->getSetting('references')['checked'] ? 'show-references' : '',
            $this->getSetting('clues')['show'] && !$this->getSetting('clues')['checked'] ? 'hide-clues' : '',
          ],
        ],
        '#cache' => [
          'tags' => $file->getCacheTags(),
        ],
      ];
    }
    return $elements;
  }

  /**
   * Return render array for crossword title.
   */
  protected function getTitle($data) {
    if ($this->getSetting('details')['title_tag']) {
      return [
        '#type' => 'html_tag',
        '#tag' => $this->getSetting('details')['title_tag'],
        '#value' => $data['title'],
        '#attributes' => [
          'class' => [
            'crossword-title',
          ],
        ],
      ];
    }
  }

  /**
   * Return render array for crossword author.
   */
  protected function getAuthor($data) {
    if ($this->getSetting('details')['author_tag'] && !empty($data['author'])) {
      return [
        '#type' => 'html_tag',
        '#tag' => $this->getSetting('details')['author_tag'],
        '#value' => $data['author'],
        '#attributes' => [
          'class' => [
            'crossword-author',
          ],
        ],
      ];
    }
  }

  /**
   * Return render array for crossword notes.
   */
  protected function getNotepad($data) {
    if ($this->getSetting('details')['notepad_tag'] && !empty($data['notepad'])) {
      return [
        '#type' => 'html_tag',
        '#tag' => $this->getSetting('details')['notepad_tag'],
        '#value' => nl2br($data['notepad']),
        '#attributes' => [
          'class' => [
            'crossword-notepad',
          ],
        ],
      ];
    }
  }

  /**
   * Return render array for across clues.
   */
  protected function getAcross($data) {
    if (isset($this->getSetting('clues')['do_not_render']) && $this->getSetting('clues')['do_not_render']) {
      return;
    }
    $render = [
      '#theme' => 'crossword_clues',
      '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
      '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
      '#content' => [],
      '#direction' => 'across',
      '#attributes' => [
        'class' => ['across'],
      ],
    ];
    if (!empty($data['puzzle']['clues']['across'])) {
      foreach ($data['puzzle']['clues']['across'] as $across_index => $across) {
        $render['#content'][] = [
          '#theme' => 'crossword_clue',
          '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
          '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
          '#text' => $across['text'],
          '#numeral' => $across['numeral'],
          '#direction' => 'across',
          '#attributes' => [
            'data-clue-index-across' => (string) $across_index,
            'data-clue-numeral-across' => $across['numeral'],
          ],
        ];
      }
    }
    return $render;
  }

  /**
   * Return render array for down clues.
   */
  protected function getDown($data) {
    if (isset($this->getSetting('clues')['do_not_render']) && $this->getSetting('clues')['do_not_render']) {
      return;
    }
    $render = [
      '#theme' => 'crossword_clues',
      '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
      '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
      '#content' => [],
      '#direction' => 'down',
      '#attributes' => [
        'class' => ['down'],
      ],
    ];
    if (!empty($data['puzzle']['clues']['down'])) {
      foreach ($data['puzzle']['clues']['down'] as $down_index => $down) {
        $render['#content'][] = [
          '#theme' => 'crossword_clue',
          '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
          '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
          '#text' => $down['text'],
          '#numeral' => $down['numeral'],
          '#direction' => 'down',
          '#attributes' => [
            'data-clue-index-down' => (string) $down_index,
            'data-clue-numeral-down' => $down['numeral'],
          ],
        ];
      }
    }
    return $render;
  }

  /**
   * Return render array for crossword grid.
   */
  protected function getGrid($data, $show_fill = FALSE) {
    $render = [
      '#theme' => 'crossword_grid',
      '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
      '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
      '#content' => [],
    ];
    if (!empty($data['puzzle']['grid'])) {
      foreach ($data['puzzle']['grid'] as $row_index => $grid_row) {
        $render_row = [
          '#theme' => 'crossword_grid_row',
          '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
          '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
          '#content' => [],
        ];
        foreach ($grid_row as $col_index => $square) {
          if ($square['fill'] === NULL) {
            $render_row['#content'][] = [
              '#theme' => 'crossword_square',
              '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
              '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
              '#attributes' => [
                'data-col' => (string) $col_index,
                'data-row' => (string) $row_index,
                'class' => [
                  'black',
                ],
              ],
            ];
          }
          else {
            $render_row['#content'][] = [
              '#theme' => 'crossword_square',
              '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
              '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
              '#fill' => $show_fill ? strtoupper($square['fill']) : NULL,
              '#numeral' => isset($square['numeral']) ? $square['numeral'] : NULL,
              '#attributes' => [
                'data-col' => (string) $col_index,
                'data-row' => (string) $row_index,
                'data-clue-index-across' => isset($square['across']['index']) ? (string) $square['across']['index'] : NULL,
                'data-clue-index-down' => isset($square['down']['index']) ? (string) $square['down']['index'] : NULL,
                'data-numeral' => isset($square['numeral']) ? $square['numeral'] : NULL,
                'data-fill' => $square['fill'],
                'data-circle' => $square['circle'],
                'data-rebus' => $square['rebus'],
              ],
            ];
          }
        }
        $render['#content'][] = $render_row;
      }
    }
    return $render;
  }

  /**
   * Return render array of crossword controls.
   */
  protected function getControls($file) {
    $controls = [
      '#theme' => 'crossword_controls',
      '#screenreader' => $this->getPluginId() == 'crossword_screenreader',
      '#pseudofield' => $this->getPluginId() == 'crossword_pseudofields',
      '#content' => [
        'clues' => $this->getShowCluesCheckbox(),
        'errors' => $this->getShowErrorsCheckbox(),
        'references' => $this->getShowReferencesCheckbox(),
        'rebus' => $this->getRebusEntryCheckbox($file),
        'cheat' => $this->getButton('cheat'),
        'solution' => $this->getButton('solution'),
        'clear' => $this->getButton('clear'),
        'undo' => $this->getButton('undo'),
        'redo' => $this->getButton('redo'),
        'instructions' => $this->getButton('instructions'),
      ],
    ];
    if (isset($this->getSetting('buttons')['buttons']['instructions']['show']) && $this->getSetting('buttons')['buttons']['instructions']['show']) {
      $controls['#attached']['library'] = ['crossword/crossword.instructions'];
    }
    return $controls;
  }

  /**
   * Return render array for clues checkbox.
   */
  protected function getShowCluesCheckbox() {
    if (isset($this->getSetting('clues')['do_not_render']) && $this->getSetting('clues')['do_not_render']) {
      return;
    }
    if ($this->getSetting('clues')['show']) {
      return [
        '#type' => 'checkbox',
        '#title' => $this->getSetting('clues')['input_label'],
        '#title_display' => 'after',
        '#label_for' => 'show-clues',
        '#attributes' => [
          'class' => [
            'show-clues',
          ],
          'checked' => $this->getSetting('clues')['checked'] ? "checked" : NULL,
          'name' => 'show-clues',
          'id' => 'show-clues',
        ],
      ];
    }
  }

  /**
   * Return render array for show errors checkbox.
   */
  protected function getShowErrorsCheckbox() {
    if ($this->getSetting('errors')['show']) {
      return [
        '#type' => 'checkbox',
        '#title' => $this->getSetting('errors')['input_label'],
        '#title_display' => 'after',
        '#label_for' => 'show-errors',
        '#attributes' => [
          'class' => [
            'show-errors',
          ],
          'checked' => $this->getSetting('errors')['checked'] ? "checked" : NULL,
          'name' => 'show-errors',
          'id' => 'show-errors',
        ],
      ];
    }
  }

  /**
   * Return render array for show references checkbox.
   */
  protected function getShowReferencesCheckbox() {
    if ($this->getSetting('references')['show']) {
      return [
        '#type' => 'checkbox',
        '#title' => $this->getSetting('references')['input_label'],
        '#title_display' => 'after',
        '#label_for' => 'show-references',
        '#attributes' => [
          'class' => [
            'show-references',
          ],
          'checked' => $this->getSetting('references')['checked'] ? "checked" : NULL,
          'name' => 'show-references',
          'id' => 'show-references',
        ],
      ];
    }
  }

  /**
   * Return render array for rebus entry checkbox.
   */
  protected function getRebusEntryCheckbox($file) {
    if ($this->getSetting('rebus')['show'] || $this->crosswordDataService->isRebus($file)) {
      return [
        '#type' => 'checkbox',
        '#title' => $this->getSetting('rebus')['input_label'],
        '#title_display' => 'after',
        '#label_for' => 'rebus-entry',
        '#attributes' => [
          'class' => [
            'rebus-entry',
          ],
          'checked' => NULL,
          'name' => 'rebus-entry',
          'id' => 'rebus-entry',
        ],
      ];
    }
  }

  /**
   * Return render array of various buttons.
   */
  protected function getButton($name) {
    if (isset($this->getSetting('buttons')['buttons'][$name]['show']) && $this->getSetting('buttons')['buttons'][$name]['show']) {
      $class = ["button-$name"];
      if (isset($this->getSetting('buttons')['class'])) {
        $custom_classes = explode(" ", $this->getSetting('buttons')['class']);
        foreach ($custom_classes as $custom_class) {
          $class[] = Html::getClass($custom_class);
        }
      }
      return [
        '#type' => 'html_tag',
        '#tag' => 'button',
        '#value' => $this->getSetting('buttons')['buttons'][$name]['input_label'],
        '#attributes' => [
          'class' => $class,
          'data-confirm' => $this->getSetting('buttons')['buttons'][$name]['confirm'] ?? NULL,
        ],
      ];
    }
  }

}
