<?php

namespace Drupal\Tests\crossword\FunctionalJavascript;

/**
 * Tests the real pseudofields, not just the formatter.
 *
 * @group crossword
 */
class CrosswordPseudofieldsTest extends CrosswordFormatterTestBase {

  /**
   * {@inheritdoc}
   */
  protected $formatter = 'crossword_pseudofields';

  /**
   * {@inheritdoc}
   */
  public static $modules = ['crossword_pseudofields_tests'];

  /**
   * {@inheritdoc}
   */
  protected $testsActiveClues = TRUE;

  /**
   * Test with no clue list.
   *
   * When there are no clue dom nodes, the active clues works differently.
   * We just need to do a quick check.
   */
  public function testWithNoClueList() {
    $config = \Drupal::service('config.factory')->getEditable('core.entity_view_display.node.crossword.default');
    $content = $config->get('content');
    unset($content['crossword_across']);
    unset($content['crossword_down']);
    $config->set('content', $content);
    $config->save();

    $node = $this->createTestNode();
    $this->assertEqual(1, $node->id());

    // View the crossword node.
    $this->drupalGet("crossword-tests/1/{$this->formatter}");
    $assertSession = $this->assertSession();
    $session = $this->getSession();
    $page = $session->getPage();

    // Confirm clues are not listed.
    $assertSession->pageTextNotContains('Across');
    $assertSession->pageTextNotContains('Down');

    // Check initial active clue.
    $session->wait(20);
    if ($this->testsActiveClues) {
      $assertSession->elementTextContains('css', '#active-clues', "3A Second square has a circle");
    }

    // Click on bottom left square. Check that active clues update.
    $square = $page->find('css', '[data-col="0"][data-row="2"]');
    $input = $page->find('css', '[data-col="0"][data-row="2"] input');
    $square->click();
    $session->wait(20);
    $this->assertTrue($square->hasClass('active'));
    if ($this->testsActiveClues) {
      $assertSession->elementTextContains('css', '#active-clues', "5A Has a reference to 3-Across and 1-Down");
      $assertSession->elementTextContains('css', '#active-clues', "3A Second square has a circle");
      $assertSession->elementTextContains('css', '#active-clues', "1D is AB2");
    }

    // Uncheck references. Make sure active clues update.
    $page->find('css', '#show-references')->click();
    $session->wait(20);
    if ($this->testsActiveClues) {
      $assertSession->elementTextContains('css', '#active-clues', "5A Has a reference to 3-Across and 1-Down");
      $assertSession->elementTextNotContains('css', '#active-clues', "3A Second square has a circle");
      $assertSession->elementTextNotContains('css', '#active-clues', "1D is AB2");
    }
  }

  /**
   * Check that special behavior with no clue list is safe.
   */
  public function testWithNoClueListXss() {
    $this->testPuzzleFilename = 'xss.txt';

    $config = \Drupal::service('config.factory')->getEditable('core.entity_view_display.node.crossword.default');
    $content = $config->get('content');
    unset($content['crossword_across']);
    unset($content['crossword_down']);
    $config->set('content', $content);
    $config->save();

    $node = $this->createTestNode();
    $this->assertEqual(1, $node->id());

    // The rest is copied from CrosswordFormatterXssTest::testCrosswordFormatter().
    // View the crossword node.
    $this->drupalGet("crossword-tests/1/{$this->formatter}");
    $assertSession = $this->assertSession();
    $session = $this->getSession();
    $page = $session->getPage();

    // Confirm clues are not listed.
    $assertSession->pageTextNotContains('Across');
    $assertSession->pageTextNotContains('Down');

    // These is attempted XSS all over this puzzle. If it causes an alert
    // to appear at any time, there would end up being an UnexpectedAlertOpen
    // and the test would err out. If we get to the end and everything passes
    // then we know no alert ever appeared.

    // Make sure the attempted Xss text is on the page.
    // These three elements should get script stripped.
    $assertSession->pageTextContains("alert('Xss Title');");
    $assertSession->pageTextContains("alert('Xss Author');");
    $assertSession->pageTextContains("alert('Xss Notepad');");

    // Click on top right square and cheat.
    // This square is un-crossed so direction changes to down.
    $square = $page->find('css', '[data-col="2"][data-row="0"]');
    $square->click();
    $page->find('css', '.button-cheat')->click();
    $session->wait(20);
    $assertSession->pageTextContains("<script>alert('Xss Letter');</script>");

    // Click (twice) on 3-Across and see clue.
    $square = $page->find('css', '[data-col="0"][data-row="1"]');
    $square->click();
    $square->click();
    $session->wait(20);
    $assertSession->elementTextContains('css', '#active-clues', "<script>alert('Xss Clue');</script>");

    // Click on 5-Across and see reference.
    $square = $page->find('css', '[data-col="0"][data-row="2"]');
    $square->click();
    $session->wait(20);
    $assertSession->elementTextContains('css', '#active-clues', "<script>alert('Xss Reference');</script>");

    // Just for fun, do an active test of alert presence.
    try {
      $session->getDriver()->getWebDriverSession()->getAlert_text();
      $exception = NULL;
    }
    catch (\Exception $e) {
      $exception = $e;
    }
    $this->assertEquals('WebDriver\Exception\NoAlertOpenError', get_class($exception));

    // Reload to examine local storage.
    $this->drupalGet("crossword-tests/1/{$this->formatter}");
    $assertSession = $this->assertSession();
    $session = $this->getSession();
    $assertSession->pageTextContains("<script>alert('Xss Letter');</script>");
    try {
      $session->getDriver()->getWebDriverSession()->getAlert_text();
      $exception = NULL;
    }
    catch (\Exception $e) {
      $exception = $e;
    }
    $this->assertEquals('WebDriver\Exception\NoAlertOpenError', get_class($exception));
  }

  /**
   * Test clue toggle.
   */
  public function testCluesToggle() {
    $node = $this->createTestNode();
    $this->assertEqual(1, $node->id());

    // View the crossword node at its normal path.
    // Allows us to test non-default config.
    $this->drupalGet('node/1');
    $assertSession = $this->assertSession();
    $session = $this->getSession();
    $page = $session->getPage();

    // By default, there is no clue toggling and clues are visible.
    $assertSession->elementNotExists('css', '#show-clues');
    $assertSession->elementExists('css', '.crossword-clues.across');
    $assertSession->elementExists('css', '.crossword-clues.down');

    // Add the toggle button, initially unchecked.
    $config = \Drupal::service('config.factory')->getEditable('crossword_pseudofields.settings');
    $config->set('clues.show', TRUE);
    $config->save();
    drupal_flush_all_caches();

    $this->drupalGet('node/1');
    $assertSession->elementExists('css', '#show-clues');
    $assertSession->elementNotExists('css', '#show-clues:checked');
    $this->assertFalse($page->find('css', '.crossword-clues.across')->isVisible());
    $this->assertFalse($page->find('css', '.crossword-clues.down')->isVisible());
    // Click the toggle button to show clues.
    $page->find('css', '#show-clues')->click();
    $session->wait(20);
    $this->assertTrue($page->find('css', '.crossword-clues.across')->isVisible());
    $this->assertTrue($page->find('css', '.crossword-clues.down')->isVisible());

    // Add the toggle button, initially checked.
    $config = \Drupal::service('config.factory')->getEditable('crossword_pseudofields.settings');
    $config->set('clues.checked', TRUE);
    $config->save();
    drupal_flush_all_caches();

    $this->drupalGet('node/1');
    $assertSession->elementExists('css', '#show-clues');
    $assertSession->elementExists('css', '#show-clues:checked');
    $this->assertTrue($page->find('css', '.crossword-clues.across')->isVisible());
    $this->assertTrue($page->find('css', '.crossword-clues.down')->isVisible());
    // Click the toggle button to show clues.
    $page->find('css', '#show-clues')->click();
    $session->wait(20);
    $this->assertFalse($page->find('css', '.crossword-clues.across')->isVisible());
    $this->assertFalse($page->find('css', '.crossword-clues.down')->isVisible());
  }

}
